#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta, abstractmethod
from typing import Callable
from contextlib import suppress
from functools import partial
import datetime
import numpy as np
from ..date_handler import global_evaluation_date
from ..utilities.enums import EngineType
from ..utilities.utility import time_this, logging
from ..pricing_engine_base.engine_base import PricingEngineBase


class OptionBase(metaclass=ABCMeta):
    """衍生品基类"""

    def __init__(self):
        """初始化
        process: StochProcessBase，标的资产随机过程
        engine: PricingEngine，定价引擎
        engine_method: str, 定价引擎种类。
                若未输入engine，可以用engine_method快捷生成指定定价引擎。可选Analytic、MC/MonteCarlo、PDE/FDM、QUAD、TREE。
        trade_calendar: TradeCalendar，交易日历
        start_date : datetime.date，起始日
        end_date: datetime.date，到期日
        """
        self.process = None
        self.engine = None
        self.trade_calendar = None
        self.start_date = None
        self.end_date = None
        # 反算波动率
        self.calc_implied_bs_vol = partial(self.calc_param, reset_param_method=self.reset_vol)

    @abstractmethod
    def set_pricing_engine(self, engine):
        """设置定价引擎，指定当前的定价函数_pricing_func.
        根据产品类型，可能存在闭式解、蒙特卡洛模拟、PDE有限差分、积分法、树方法等定价方法。
        """

    @abstractmethod
    def __repr__(self):
        """返回期权的描述"""

    @abstractmethod
    def price(self, *args, **kwargs):
        """执行定价
        对于标的期初价格为s0的期权，以香草期权为例，返回的是BSM公式的定价结果，即max(0, St - K)期望的现值
        如果想要得到单位估值，需要再除以s0"""
        self.validate_parameters()

    def validate_parameters(self, t=None):
        """price方法中，先检查产品是否已经配置了定价引擎，估值日是否小于终止日
        Args:
            t: datetime.date，计算期权价格的日期，默认None，此时使用全局估值日"""
        if self.engine is None or not isinstance(self.engine, PricingEngineBase):
            raise AttributeError("尚未配置定价引擎，您可以在以下4种方法中选择:\n"
                                 "    1. 在产品类同时输入s,r,q,vol以创建定价引擎，可以用engine_method='MC'/'PDE'等指定引擎种类;\n"
                                 "    2. 在产品类同时输入s,r,q,vol和engine_method='MC'/'PDE'...以创建指令种类的定价引擎;\n"
                                 "    3. 在产品类同时输入标的资产随机过程StochProcess实例和engine_method='MC'/'PDE'...以创建指令种类的定价引擎;\n"
                                 "    4. 自行创建定价引擎，再使用set_pricing_engine方法配置到产品中")
        calculate_date = global_evaluation_date() if t is None else t
        date_message = None
        if calculate_date < self.start_date:
            date_message = f"估值日必须大于等于起始日, 当前估值日为{calculate_date}, 起始日为{self.start_date}.\n"
        if hasattr(self, "stop_date") and calculate_date > self.stop_date:  # 某些带有敲出的衍生品，会有终止日stop_date属性
            date_message = f"估值日必须小于等于终止日, 当前估值日为{calculate_date}, 终止日为{self.stop_date}.\n"
        elif calculate_date > self.end_date:
            date_message = f"估值日必须小于等于到期日, 当前估值日为{calculate_date}, 到期日为{self.end_date}.\n"
        if date_message:
            raise ValueError(date_message + f"默认估值日是距离今天最近的交易日。请在price(t=evaluation_date)指定估值日，或者使用set_evaluation_date(datetime.date(YYYY, mm, dd))配置全局估值日。")

    def delta(self, spot=None, step=None, *args, **kwargs):
        """求价格spot处的delta值
        会先尝试调用定价引擎的delta方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的0.1%
        Returns: delta = ∂V/∂S
        """
        try:
            return self.engine.delta(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            step = spot * 0.001 if step is None else step
            up_price = self.price(spot=spot + step)
            down_price = self.price(spot=spot - step)
            delta = (up_price - down_price) / (2 * step)
            return delta

    def gamma(self, spot=None, step=None, *args, **kwargs):
        """求价格spot处的gamma值
        会先尝试调用定价引擎的gamma方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的0.1%
        Returns: gamma = ∂2V/∂S2
        """
        try:
            return self.engine.gamma(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            step = spot * 0.001 if step is None else step
            up_price = self.price(spot=spot + step)
            down_price = self.price(spot=spot - step)
            mid_price = self.price(spot=spot)
            gamma = (up_price - 2 * mid_price + down_price) / (step ** 2)
            return gamma

    def vega(self, spot=None, step=0.01, *args, **kwargs):
        """计算波动率上升1%的vega todo:目前只支持常数波动率
        会先尝试调用定价引擎的vega方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: vega = ∂V/∂σ
        """
        try:
            return self.engine.vega(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            last_vol = self.engine.process.vol.data
            last_price = self.price(spot=spot)
            self.engine.process.vol.data = last_vol + step
            new_price = self.price(spot=spot)
            self.engine.process.vol.data = last_vol
            vega = (new_price - last_price)
            return vega

    def theta(self, spot=None, step=1, *args, **kwargs):
        """计算一天的theta值
        会先尝试调用定价引擎的theta方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: int，时间步长，默认为1天
        Returns: theta = ∂V/∂t
        """
        try:
            return self.engine.theta(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            try:
                spot = self.engine.process.spot() if spot is None else spot
                step = datetime.timedelta(days=step)
                origin_price = self.price(spot=spot)
                next_date = self.trade_calendar.advance(global_evaluation_date(), step)
                new_price = self.price(t=next_date, spot=spot)
                theta = (new_price - origin_price)
                return theta
            except:  # 可能遇到到期日估值，无法计算theta
                calculate_date = global_evaluation_date()
                if calculate_date > self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1)):
                    raise ValueError(
                        f"估值日必须小于等于到期日, 当前估值日为{calculate_date}, 到期日为{self.end_date}.")
                return np.nan  # 到期日估值，无法计算theta,返回nan

    def rho(self, spot=None, step=0.01, *args, **kwargs):
        """计算无风险利率上升1%的rho todo:目前只支持常数无风险利率
        会先尝试调用定价引擎的rho方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: rho = ∂V/∂r
        """
        try:
            return self.engine.rho(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            last_r = self.engine.process.interest.data
            last_price = self.price(spot=spot)
            self.engine.process.interest.data = last_r + step
            new_price = self.price(spot=spot)
            self.engine.process.interest.data = last_r
            rho = (new_price - last_price)
            return rho

    def vanna(self, spot=None, step=0.001, *args, **kwargs):
        """计算vanna todo:目前只支持常数波动率
        会先尝试调用定价引擎的vanna方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: vanna = ∂2V/∂S∂σ
        """
        try:
            return self.engine.vanna(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            last_vol = self.engine.process.vol.data
            last_delta = self.delta(spot=spot)
            self.engine.process.vol.data = last_vol + step
            new_delta = self.delta(spot=spot)
            self.engine.process.vol.data = last_vol
            vanna = (new_delta - last_delta) / step
            return vanna

    def volga(self, spot=None, step=0.001, *args, **kwargs):
        """计算volga todo:目前只支持常数波动率
        会先尝试调用定价引擎的volga方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: volga = ∂2V/∂σ2
        """
        try:
            return self.engine.volga(prod=self, spot=spot, step=step, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            last_vol = self.engine.process.vol.data
            price = self.price(spot=spot)
            self.engine.process.vol.data = last_vol - step
            down_price = self.price(spot=spot)
            self.engine.process.vol.data = last_vol + step
            up_price = self.price(spot=spot)
            self.engine.process.vol.data = last_vol
            volga = (up_price - 2 * price + down_price) / step ** 2
            return volga

    def pv_and_greeks(self, spot=None, *args, **kwargs):
        """当一次性计算pv和5个greeks时，可以调用此函数，减少计算量
        会先尝试调用定价引擎的pv_and_greeks方法，如果没有则使用通用的差分法计算
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
        Returns:
            Dict[str:float]: {'pv': pv, 'delta': delta, 'gamma': gamma, 'theta': theta, 'vega': vega, 'rho': rho}
        """
        # if hasattr(self.engine, "pv_and_greeks"):
        #     with suppress(AttributeError, ValueError):
        try:
            if self.engine.engine_type == EngineType.PdeEngine:  # PDE定价引擎需要先经过一次定价，生成S-t估值矩阵
                self.price()
            return self.engine.pv_and_greeks(prod=self, spot=spot, *args, **kwargs)
        except:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            pv = np.float64(self.price(spot=spot))
            s_step = spot * 0.01
            up_price = self.price(spot=spot + s_step)
            down_price = self.price(spot=spot - s_step)
            delta = (up_price - down_price) / (2 * s_step)
            gamma = (up_price - 2 * pv + down_price) / (s_step ** 2)

            last_vol = self.engine.process.vol.data
            self.engine.process.vol.data = last_vol + 0.01
            new_price = self.price(spot=spot)
            self.engine.process.vol.data = last_vol
            vega = (new_price - pv)

            last_r = self.engine.process.interest.data
            self.engine.process.interest.data = last_r + 0.01
            new_price = self.price(spot=spot)
            self.engine.process.interest.data = last_r
            rho = (new_price - pv)

            try:  # 可能遇到到期日估值，无法计算theta
                t_step = datetime.timedelta(days=1)
                next_date = self.trade_calendar.advance(global_evaluation_date(), t_step)
                new_price = self.price(t=next_date, spot=spot)
                theta = (new_price - pv)
            except Exception as e:
                logging.error(e)
                theta = np.nan
            return {'pv': pv, 'delta': delta, 'gamma': gamma, 'vega': vega, 'theta': theta, 'rho': rho}

    @time_this
    def calc_param(self, reset_param_method: Callable, target_value: float = None, t: datetime.date = None, spot=None,
                   high=2., low=0., thres=0.001, max_iters=20, *args, **kwargs):
        """根据目标估值Y，用二分法反算衍生品参数X
        Args:
            reset_param_method: class method，重置衍生品参数X的类方法
            target_value: float，目标估值，对期权而言是期权的市场价格(期权费)，对票据是估值。
                                票据的公允估值=名义本金(pricelib估值不含名本，因此默认target_value=期初价格s0)
            t: datetime.date，计算衍生品价格的日期
            spot: float，计算衍生品价格的标的资产价格
            high: float，二分法的衍生品参数X搜索区间高点
            low: float，二分法的衍生品参数X搜索区间低点
            thres: float，判断二分法收敛的阈值
            max_iters: int, 二分法最大迭代次数，避免无限循环
        Returns:
        """
        mid, estimate = None, None
        for i in range(max_iters):
            mid = (high + low) / 2
            if mid < thres:
                mid = thres
            temp_param = mid
            reset_param_method(temp_param, *args, **kwargs)
            estimate = self.price(t=t, spot=spot)
            logging.info(f"{reset_param_method.__name__}参数 = {temp_param}: estimate = {estimate}")
            if abs(estimate - target_value) <= abs(thres * target_value):
                break
            elif estimate > target_value * (1 + thres):
                high = mid
            elif estimate < target_value * (1 - thres):
                low = mid
        return mid, estimate

    def reset_vol(self, vol):
        """重置标的资产的波动率的值"""
        self.process.volatility.data = vol



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
import numpy as np
from pricelib.common.utilities.enums import TouchType, EngineType, ExerciseType, PaymentType, StatusType
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer, global_auto_update_status
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.analytic_engines import AnalyticDoubleDigitalEngine
from pricelib.pricing_engines.fdm_engines import FdmDigitalEngine
from pricelib.pricing_engines.mc_engines import MCDoubleDigitalEngine


class DoubleDigitalOption(OptionBase, Observer):
    """双边二元期权
        欧式: 二元凹式/二元凸式
        美式: 美式双接触/美式双不接触"""

    def __init__(self, touch_type=TouchType.Touch, exercise_type=ExerciseType.American, payment_type=PaymentType.Expire,
                 *, strike_lower=95., strike_upper=105., rebate_lower=0., rebate_upper=0., rebate_annual=False,
                 status=StatusType.NoTouch,maturity=None, start_date=None, end_date=None, discrete_obs_interval=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243, engine=None,
                 stoch_process=None, s0=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """构造函数
        产品参数:
            touch_type: 接触方式 - 接触Touch/不接触NotTouch
            exercise_type: 行权方式，ExerciseType枚举类，European/American
            payment_type: 支付方式，PaymentType枚举类，立即支付Hit/到期支付Expire
            bound: (strike_lower, strike_upper), (低行权价/高行权价)
            rebate: (rebate_lower, rebate_upper), 行权收益的绝对数值，非年化
                    双接触: (下边界payoff, 上边界payoff); 双不接触: (到期不接触的payoff, 无作用)
            rebate_annual: bool, 现金返还是否为年化，默认为False，即绝对数值，非年化，非百分比；若为True，则为百分比年化收益率
            status: 定价时，触碰障碍的状态，StatusType枚举类，默认为NoTouch未触碰，UpTouch为向上触碰，DownTouch为向下触碰
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
            discrete_obs_interval: 观察时间间隔. 若为连续观察，None；若为均匀离散观察，为年化的观察时间间隔
        可选参数:
            s0: float，标的期初价格，只有当 rebate_annual = True，即现金返还为年化收益率时；
                       需要输入标的期初价格，将百分比的收益率转化成绝对数值
            engine: 定价引擎，PricingEngine类
                    解析解: AnalyticDoubleDigitalEngine - Hui(1996) 级数近似解，双接触期权的下边界payoff和上边界payoff必须相等
                                                        支持 连续/离散观察、到期支付的(美式)双接触/双不接触
                    PDE: FdmDigitalEngine 支持 (欧式)二元凹式/二元凸式；支持 连续/离散观察、立即/到期支付的(美式)双接触/美式双不接触
                    蒙特卡洛: MCDoubleDigitalEngine 支持 (欧式)二元凹式/二元凸式；只支持离散观察、立即/到期支付的(美式)双接触/美式双不接触
            engine_method: str, 定价引擎种类。若未输入engine，可以用engine_method快捷生成指定定价引擎。可选Analytic、MC/MonteCarlo、PDE/FDM。
            若未提供引擎engine、也未指定引擎种类engine_method的情况下，如果提供了 随机过程 或 标的价格、无风险利率、分红/融券率、波动率，
            则使用默认定价引擎：
                到期支付的美式双接触/双不接触默认使用 Hui(1996) 级数近似解 AnalyticDoubleDigitalEngine
                其余情况默认使用 PDE 定价引擎 FdmDigitalEngine
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if self.trade_calendar.isholiday(self.end_date):  # 如果到期日不是交易日，向后顺延一个交易日
            self.end_date = self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1))
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        assert strike_lower < strike_upper, "ValueError: 双边二元期权的低行权价必须小于高行权价"
        self.bound = (strike_lower, strike_upper)  # (低行权价, 高行权价)
        self.rebate = (rebate_lower, rebate_upper)  # 双接触: (下边界payoff, 上边界payoff); 双不接触: (不接触payoff, 无作用)
        self.rebate_annual = rebate_annual  # 现金返还是否为年化
        self.touch_type = touch_type
        self.exercise_type = exercise_type
        self.payment_type = payment_type
        if self.exercise_type == ExerciseType.European or self.touch_type == TouchType.NoTouch:
            assert payment_type == PaymentType.Expire, "ValueError: 欧式双边二元期权或美式双不触碰期权的现金返还方式PaymentType一定是到期支付Expire"
        self.discrete_obs_interval = discrete_obs_interval  # 连续观察=None；均匀离散观察=观察时间间隔
        if self.exercise_type == ExerciseType.European:
            assert self.discrete_obs_interval is None, "ValueError: 欧式双边二元期权，观察时间间隔只能为None."
        self.s0 = s0
        self.status = status
        self.stop_date = self.end_date  # 停止日期，默认等于到期日。发生提前行权，则停止日期更新为提前行权日期
        if engine is not None:
            self.set_pricing_engine(engine)
        else:
            if engine_method is None:
                if (exercise_type == ExerciseType.European or (  # 支持二元凸式/二元凹式解析解
                        exercise_type == ExerciseType.American and payment_type == PaymentType.Expire and (
                        touch_type == TouchType.NoTouch or (  # 美式双接触/双不接触只支持到期支付，
                        touch_type == TouchType.Touch and rebate_lower == rebate_upper)))):  # 若为美式双接触，则上下边界payoff必须相等
                    default_engine = AnalyticDoubleDigitalEngine
                else:
                    default_engine = FdmDigitalEngine
            elif engine_method == "Analytic":
                default_engine = AnalyticDoubleDigitalEngine
            elif engine_method in ["PDE", "FDM"]:
                default_engine = FdmDigitalEngine  # todo 支持 rebate_annual 年化的现金返还
            elif engine_method in ["MC", "MonteCarlo"]:
                default_engine = MCDoubleDigitalEngine  # todo 支持 rebate_annual 年化的现金返还
            else:
                raise NotImplementedError(
                    f"不支持的engine_method{engine_method}, {self.__repr__()}仅支持Analytic/MC/PDE")
            if stoch_process is not None:
                self.set_pricing_engine(default_engine(stoch_process))
            elif s is not None and r is not None and q is not None and vol is not None:
                self.set_pricing_engine(default_engine(s=s, r=r, q=q, vol=vol))



    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经触碰障碍"""
        if global_evaluation_date() > self.stop_date:  # 如果估值日已经过了终止日，维持原来的状态
            return
        if global_auto_update_status() and observable == self.engine.process.spot:
            # 如果是欧式双边二元，且估值日是到期日时，更新状态  todo: 暂时只支持日频行情更新，即估值日的第一次标的价格变化即视为收盘价更新
            if self.exercise_type == ExerciseType.European and global_evaluation_date() == self.end_date:
                self.change_status(observable.data)
            # 如果是美式双边二元，连续观察或每日离散观察，每次标的价格变化，均更新状态
            elif self.exercise_type == ExerciseType.American:
                # todo: 均匀离散观察，只支持每日观察，暂不支持其他观察频率。即每次标的价格变化，均更新状态
                self.change_status(observable.data)

    def change_status(self, spot):
        """获取状态变化"""
        if self.status == StatusType.NoTouch:
            if spot >= self.bound[1]:
                self.status = StatusType.UpTouch
                if self.payment_type == PaymentType.Hit:  # 若美式双接触提前终止
                    self.stop_date = global_evaluation_date()
            elif spot <= self.bound[0]:
                self.status = StatusType.DownTouch
                if self.payment_type == PaymentType.Hit:  # 若美式双接触提前终止
                    self.stop_date = global_evaluation_date()
        else:
            return

    def __repr__(self):
        """返回期权的描述"""
        if self.exercise_type == ExerciseType.European:
            if self.touch_type == TouchType.Touch:
                return f"二元凹式(欧式双接触期权)"
            else:  # NoTouch
                return f"二元凸式(欧式双不接触期权)"
        elif self.exercise_type == ExerciseType.American:
            if self.discrete_obs_interval is None:
                obs_type = "连续观察"
            else:
                obs_type = "离散观察"
            return f"{self.payment_type.value}{obs_type}美式双{self.touch_type.value}二元期权"
        else:
            raise ValueError("无法识别的期权类型")

    @time_this
    def price(self, t=None, spot=None):
        self.validate_parameters(t=t)
        if self.status != StatusType.NoTouch:  # 如果已经触碰障碍，则直接返回行权收益的现值
            if self.touch_type == TouchType.NoTouch:
                return 0
            elif self.touch_type == TouchType.Touch:
                calculate_date = global_evaluation_date() if t is None else t
                if self.payment_type == PaymentType.Hit:  # 立即支付
                    if self.rebate_annual:  # 年化百分比
                        elapsed_time = (calculate_date - self.start_date).days / self.annual_days.value
                        _rebate = [r * self.s0 * elapsed_time for r in self.rebate]
                    else:  # 绝对数值
                        _rebate = self.rebate
                    if self.status == StatusType.UpTouch:
                        return _rebate[1]
                    elif self.status == StatusType.UpTouch:
                        return _rebate[0]
                elif self.payment_type == PaymentType.Expire:  # 到期支付
                    if self.rebate_annual:  # 年化百分比
                        tenor = (self.end_date - self.start_date).days / self.annual_days.value  # 从起始日到终止日的自然日时间（年）
                        _rebate = [r * self.s0 * tenor for r in self.rebate]
                    else:  # 绝对数值
                        _rebate = self.rebate
                    _maturity = (self.end_date - calculate_date).days / self.annual_days.value
                    if self.status == StatusType.UpTouch:
                        return _rebate[1] * self.engine.process.interest.disc_factor(_maturity)
                    elif self.status == StatusType.UpTouch:
                        return _rebate[0] * self.engine.process.interest.disc_factor(_maturity)
                else:
                    raise ValueError(f"未知的支付方式{self.payment_type}")
            else:
                raise ValueError("Error: TouchType只能是NoTouch或Touch")
        else:  # 尚未触碰障碍，正常定价
            if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边障碍
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=self.bound, rebate=self.rebate)
            else:
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from .fdm_barrier_engine import FdmBarrierEngine


class FdmAirbagEngine(FdmBarrierEngine):
    """安全气囊PDE有限差分法定价引擎
    安全气囊是一个向下敲入的期权，到期支付，敲入前是call或牛市价差，敲入后的收益是标的涨跌幅
    如果存在敲入前或敲入后封顶收益率，或者封底收益率，则需要改写终止条件和边界条件
    支持连续观察/离散观察(默认为每日观察)"""

    def _init_boundary_condition(self, smax, maturity, elapsed_time=None, _floor_yield=0):
        """初始化边界条件
        安全气囊，未敲入是call，已敲入是标的资产
        Args:
            smax: float, 有限差分网格上边界的价格最大值
            maturity: float, 到期时间
        Returns:
            self.fn_callput: List[Callable], 0~smax网格的边界条件
            self.fn_bound: List[Callable], 障碍价格网格的边界条件
        """
        # 生成已敲入网格的边界条件： 0~smax，此时收益为标的资产涨跌幅（可能存在敲入后封顶收益率，或者封底收益率）
        if self.prod.in_strike_floor is not None:  # 若存在敲入后封底价格
            _in_min_s = self.prod.in_strike_floor
        else:
            _in_min_s = 0
        if self.prod.in_strike_cap is not None:  # 若存在敲入后封顶价格
            _in_max_s = self.prod.in_strike_cap
        else:
            _in_max_s = smax
        self.fn_callput = [lambda u: (self.prod.knockin_parti * (_in_min_s - self.prod.s0) + _floor_yield
                                      ) * self.process.interest.disc_factor(maturity, u),
                           lambda u: (self.prod.reset_call_parti * (_in_max_s - self.prod.s0) + _floor_yield
                                      ) * self.process.interest.disc_factor(maturity, u)]
        # 生成未敲入网格的边界条件： bound[0]~smax, 此时收益为向下敲出看涨期权或者价差
        if self.prod.init_strike_cap is not None:  # 若存在敲入后封顶价格
            _max_s = self.prod.init_strike_cap
        else:
            _max_s = smax
        # 返回一个包含两个lambda函数的列表，第u天触碰障碍产生的payoff，折现到第u天的价值。 列表索引分别为[0, 1]，每个lambda函数的形参r=self.rebate中的r补偿值
        self.fn_bound = []
        for i, _bound in enumerate(self.bound):
            if _bound in [-np.inf, np.inf]:  # 安全气囊没有上方的障碍，即障碍价无穷。未敲入call的上边界是smax（若封顶则为init_strike_cap）
                self.fn_bound.append(lambda u: (self.prod.call_parti * (_max_s - self.prod.strike) + _floor_yield
                                                ) * self.process.interest.disc_factor(maturity, u))
            else:  # 安全气囊有下方的敲入障碍价格，触碰下方的障碍后，到期时支付，安全气囊已敲入是标的资产跌幅
                if self.prod.discrete_obs_interval is None:  # 连续观察, 价格下边界是self.bound[0]
                    self.fn_bound.append(
                        lambda u: (self.prod.knockin_parti * (max(_bound, _in_min_s) - self.prod.s0)
                                   + _floor_yield) * self.process.interest.disc_factor(maturity, u))
                else:  # 离散观察，价格下边界是0
                    self.fn_bound.append(
                        lambda u: (self.prod.knockin_parti * (_in_min_s - self.prod.s0) + _floor_yield
                                   ) * self.process.interest.disc_factor(maturity, u))

    def _init_terminal_condition(self, s_vec, maturity, elapsed_time=None, _floor_yield=0.):
        """初始化终止时已敲入网格终止时的期权价值，安全气囊敲入后是标的资产涨跌幅
        _init_no_touch_terminal_condition返还未敲入网格的终止条件
        连续观察：网格上下界是障碍价格
        离散观察/到期观察：网格上下界是0和smax
        """
        prod = self.prod
        # 已敲入，且标的上涨
        if prod.in_strike_cap is None:
            in_payoff_upper = s_vec - prod.s0
        else:  # 若存在敲入后封顶价格
            in_payoff_upper = np.where(s_vec - prod.in_strike_cap > 0,
                                       prod.in_strike_cap - prod.s0, s_vec - prod.s0)
        # 已敲入，且标的下跌
        if prod.in_strike_floor is None:
            in_payoff_lower = s_vec - prod.s0
        else:  # 若存在敲入后封底价格
            in_payoff_lower = np.where(s_vec - prod.in_strike_floor < 0,
                                       prod.in_strike_floor - prod.s0, s_vec - prod.s0)
        yv = (np.where(s_vec - prod.s0 > 0,
                       # 敲入且标的上涨
                       in_payoff_upper * prod.reset_call_parti,  # 敲入重置后看涨参与率
                       # 敲入且标的下跌
                       in_payoff_lower * prod.knockin_parti)  # 敲入下跌参与率
              + _floor_yield)  # 保底收益
        return yv

    def _init_no_touch_terminal_condition(self, s_vec, p=0, maturity=0, elapsed_time=None, _floor_yield=0):
        """初始化安全气囊未敲入网格终点的期权价值，
        连续观察：网格上下界是障碍价格，因此都在敲入价格以上，只有未敲入的call（或牛市价差）
        离散观察/到期观察：网格上下界是0和smax，敲入价上方，未敲入是call（或牛市价差）；敲入价下方是标的资产跌幅（或熊市价差）"""
        prod = self.prod
        # 未敲入，期末价值是call或牛市价差
        if prod.init_strike_cap is None:
            init_payoff = s_vec - prod.strike
        else:  # 若存在敲入前封顶价格
            init_payoff = np.where(s_vec - prod.init_strike_cap > 0,
                                   prod.init_strike_cap - prod.strike, s_vec - prod.strike)
        # 已敲入，且标的下跌
        if prod.in_strike_floor is None:
            in_payoff_lower = s_vec - prod.s0
        else:  # 若存在敲入后封底价格
            in_payoff_lower = np.where(s_vec - prod.in_strike_floor < 0,
                                       prod.in_strike_floor - prod.s0, s_vec - prod.s0)
        if self.prod.discrete_obs_interval is None:  # 连续观察
            yv = (_floor_yield + np.maximum(init_payoff, 0) * self.prod.call_parti  # 未敲入的看涨参与率
                  ) * self.process.interest.disc_factor(maturity, p)
        else:  # 离散观察
            yv = np.where(s_vec - self.bound[0] > 0,
                          np.maximum(init_payoff, 0) * self.prod.call_parti,  # 未敲入的看涨参与率
                          in_payoff_lower * self.prod.knockin_parti  # 敲入的下跌参与率
                          ) + _floor_yield  # 保底收益
        return yv



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
import numpy as np
from pricelib.common.utilities.enums import CallPut, TouchType, ExerciseType, PaymentType, EngineType, StatusType
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer, global_auto_update_status
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.analytic_engines import AnalyticCashOrNothingEngine
from pricelib.pricing_engines.mc_engines import MCDigitalEngine
from pricelib.pricing_engines.fdm_engines import FdmDigitalEngine
from pricelib.pricing_engines.integral_engines import QuadDigitalEngine
from pricelib.pricing_engines.tree_engines import BiTreeDigitalEngine


class DigitalOption(OptionBase, Observer):
    """单边二元(数字)期权-现金或无(cash or nothing)
    包括欧式二元(现金或无)、美式二元(立即支付)、一触即付(到期支付)，支持M. Broadie, P. Glasserman, S.G. Kou(1999)离散观察调整"""
    touch_type = TouchType.Touch  # 触碰类型：触碰

    def __init__(self, strike, rebate, callput=CallPut.Call, exercise_type=ExerciseType.European, rebate_annual=False,
                 payment_type=PaymentType.Expire, *, maturity=None, start_date=None, end_date=None, engine=None,
                 status=StatusType.NoTouch, discrete_obs_interval=None, trade_calendar=CN_CALENDAR,
                 annual_days=AnnualDays.N365, t_step_per_year=243, stoch_process=None, s0=None, s=None, r=None, q=None,
                 vol=None, engine_method=None):
        """构造函数
        产品参数:
            strike:  float, 执行价
            rebate:  float, 行权收益, 绝对数值（非年化）
            rebate_annual: bool, 现金返还是否为年化，默认为False，即绝对数值，非年化，非百分比；若为True，则为百分比年化收益率
            callput: 看涨看跌类型，CallPut枚举类，Call/Put
            exercise_type: 行权方式，ExerciseType枚举类，European/American
            payment_type: 支付方式，PaymentType枚举类，立即支付Hit/到期支付Expire
            status: 触碰障碍的状态，StatusType枚举类，默认为NoTouch未触碰，UpTouch为向上触碰，DownTouch为向下触碰
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
            discrete_obs_interval: 观察时间间隔. 若为连续观察，None；若为均匀离散观察，为年化的观察时间间隔
        可选参数:
            s0: float，标的期初价格，只有当 rebate_annual = True，即现金返还为年化收益率时；
               需要输入标的期初价格，将百分比的收益率转化成绝对数值
            engine: 定价引擎，PricingEngine类，以下几种引擎均支持欧式二元(现金或无)、美式二元(立即支付)、一触即付(到期支付)，
                            解析解和PDE引擎支持离散观察/连续观察，其余引擎只支持离散观察(默认为每日观察)
                    解析解: AnalyticCashOrNothingEngine
                    蒙特卡洛: MCDigitalEngine  todo 支持年化现金收益 rebate_annual
                    PDE: FdmDigitalEngine  todo 支持年化现金收益 rebate_annual
                    积分法: QuadDigitalEngine  todo 支持年化现金收益 rebate_annual
                    二叉树: BiTreeDigitalEngine  todo 支持年化现金收益 rebate_annual
            engine_method: str, 定价引擎种类。若未输入engine，可以用engine_method快捷生成指定定价引擎。可选Analytic、MC/MonteCarlo、PDE/FDM、QUAD、TREE。
            若未提供引擎engine、也未指定引擎种类engine_method的情况下，如果提供了 随机过程 或 标的价格、无风险利率、分红/融券率、波动率，
            则默认使用 二元(数字)期权-现金或无(cash or nothing)-闭式解定价引擎 Reiner and Rubinstein(1991b)
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if self.trade_calendar.isholiday(self.end_date):  # 如果到期日不是交易日，向后顺延一个交易日
            self.end_date = self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1))
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.strike = strike
        self.rebate = rebate  # 行权收益，绝对值/百分比年化
        self.rebate_annual = rebate_annual  # 现金返还是否为年化
        self.callput = callput
        self.exercise_type = exercise_type
        self.payment_type = payment_type
        if self.exercise_type == ExerciseType.European or self.touch_type == TouchType.NoTouch:
            assert payment_type == PaymentType.Expire, "ValueError: 欧式二元期权或美式不触碰二元期权的现金返还方式PaymentType一定是到期支付Expire"
        self.discrete_obs_interval = discrete_obs_interval  # 连续观察=None；均匀离散观察=观察时间间隔
        # if self.exercise_type == ExerciseType.European:
        #     assert self.discrete_obs_interval is None, "ValueError: 欧式二元期权，观察时间间隔只能为None." # 其实没影响
        self.s0 = s0
        self.status = status
        self.stop_date = self.end_date  # 停止日期，默认等于到期日。美式二元如果提前行权，停止日期更新为提前行权日期
        if engine is not None:
            self.set_pricing_engine(engine)
        else:
            if engine_method is None:
                default_engine = AnalyticCashOrNothingEngine  # 默认Analytic定价引擎
            elif engine_method == "Analytic":
                default_engine = AnalyticCashOrNothingEngine
            elif engine_method in ["PDE", "FDM"]:
                default_engine = FdmDigitalEngine
            elif engine_method in ["MC", "MonteCarlo"]:
                default_engine = MCDigitalEngine
            elif engine_method == "QUAD":
                default_engine = QuadDigitalEngine
            elif engine_method == "TREE":
                default_engine = BiTreeDigitalEngine
            else:
                raise NotImplementedError(
                    f"不支持的engine_method{engine_method}, {self.__repr__()}仅支持Analytic/MC/PDE/QUAD/TREE")
            if stoch_process is not None:
                self.set_pricing_engine(default_engine(stoch_process))
            elif s is not None and r is not None and q is not None and vol is not None:
                self.set_pricing_engine(default_engine(s=s, r=r, q=q, vol=vol))

    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经触碰障碍"""
        if global_evaluation_date() > self.stop_date:  # 如果估值日已经过了终止日，维持原来的状态
            return
        if global_auto_update_status() and observable == self.engine.process.spot:
            # 如果是二元欧式，且估值日是到期日时，更新状态 todo: 暂时只支持日频行情更新，即估值日的第一次标的价格变化即视为收盘价更新
            if self.exercise_type == ExerciseType.European and global_evaluation_date() == self.end_date:
                self.change_status(observable.data)
            # 二元美式/一触即付，连续观察或每日离散观察，每次标的价格变化，均更新状态
            elif self.exercise_type == ExerciseType.American:
                # todo: 均匀离散观察，只支持每日观察，暂不支持其他观察频率。即每次标的价格变化，均更新状态
                self.change_status(observable.data)

    def change_status(self, spot):
        """获取状态变化"""
        if self.status == StatusType.NoTouch:
            if self.callput == CallPut.Call and spot >= self.strike:
                self.status = StatusType.UpTouch
                if self.payment_type == PaymentType.Hit:  # 若美式二元提前终止
                    self.stop_date = global_evaluation_date()
            elif self.callput == CallPut.Put and spot <= self.strike:
                self.status = StatusType.DownTouch
                if self.payment_type == PaymentType.Hit:  # 若美式二元提前终止
                    self.stop_date = global_evaluation_date()
        else:
            return

    def __repr__(self):
        """返回期权的描述"""
        if self.exercise_type == ExerciseType.European:
            return f"欧式二元{self.callput.name}期权"
        elif self.exercise_type == ExerciseType.American:
            if self.discrete_obs_interval is None:
                obs_type = "连续观察"
            else:
                obs_type = "离散观察"
            if self.payment_type == PaymentType.Hit:
                digital_option = "美式二元(立即支付)"
            else:
                digital_option = "一触即付(到期支付)"
            return f"{obs_type}{digital_option}{self.callput.name}期权"
        else:
            raise ValueError("二元（数字）期权的exercise_type参数只支持欧式或美式")


    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        calculate_date = global_evaluation_date() if t is None else t
        if self.status != StatusType.NoTouch:  # 如果已经触碰障碍，则直接返回行权收益的现值
            if self.touch_type == TouchType.NoTouch:
                return 0  # 单边二元属性固定为Touch，这里走不到
            elif self.touch_type == TouchType.Touch:
                if self.payment_type == PaymentType.Hit:  # 立即支付
                    if self.rebate_annual:  # 年化百分比
                        elapsed_time = (calculate_date - self.start_date).days / self.annual_days.value
                        _rebate = self.rebate * self.s0 * elapsed_time
                    else:  # 绝对数值
                        _rebate = self.rebate
                    return self.rebate
                elif self.payment_type == PaymentType.Expire:  # 到期支付
                    if self.rebate_annual:  # 年化百分比
                        tenor = (self.end_date - self.start_date).days / self.annual_days.value  # 从起始日到终止日的自然日时间（年）
                        _rebate = self.rebate * self.s0 * tenor
                    else:  # 绝对数值
                        _rebate = self.rebate
                    _maturity = (self.end_date - calculate_date).days / self.annual_days.value
                    return _rebate * self.engine.process.interest.disc_factor(_maturity)
                else:
                    raise ValueError(f"未知的支付方式{self.payment_type}")
        else:  # 尚未触碰障碍，正常定价
            if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边二元
                if self.callput == CallPut.Call:
                    bound = (None, self.strike)
                else:  # Put
                    bound = (self.strike, None)
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=bound,
                                                       rebate=(self.rebate, self.rebate))
            else:
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.date_handler import global_evaluation_date
from pricelib.common.utilities.enums import InOut, UpDown, PaymentType, ExerciseType
from pricelib.common.pricing_engine_base import McEngine


class MCBarrierEngine(McEngine):
    """障碍期权 Monte Carlo 模拟定价引擎
    只支持离散观察(默认为每日观察)；敲入现金返还为到期支付；敲出现金返还为到期支付"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
                 如果输入spot = np.ndarray一维价格向量，t=终止日，则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date,
                                                                            prod.end_date)  # 从估值日到到期日的交易日时间（天）
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value  # 从估值日到到期日的自然日时间（年）
        _elapsed_days = prod.trade_calendar.business_days_between(prod.start_date, calculate_date)  # 从起始日到估值日的交易日时间（天）

        if spot is None:
            spot = self.process.spot()
        # 当估值日就是终止日时，直接用spot计算payoff，无需随机数矩阵
        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            paths = np.array([spot, ])  # spot是一个np.ndarray一维价格向量，返回盈亏测算向量，即每个价格对应的现值，而不是平均值
            _tenor = (prod.end_date - prod.start_date).days / prod.annual_days.value  # 从起始日到到期日的自然日时间（年）
        elif calculate_date == prod.stop_date:
            paths = np.array([[spot, ]])  # spot是一个数值，计算期末payoff
            _tenor = (prod.stop_date - prod.start_date).days / prod.annual_days.value  # 从起始日到终止日的自然日时间（年）
        else:
            if spot != self.process.spot():
                self.reset_paths_flag()  # 重置路径标志位，重新生成路径
            paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                        t_step_per_year=prod.t_step_per_year).copy()
            _tenor = (prod.end_date - prod.start_date).days / prod.annual_days.value  # 从起始日到到期日的自然日时间（年）

        if prod.obs_type == ExerciseType.European:  # 如果是欧式障碍期权（到期观察）
            value = self._calc_european_barrier(prod, paths, _maturity, _tenor)
        elif prod.obs_type == ExerciseType.American:  # 如果是美式障碍期权
            value = self._calc_american_barrier(prod, paths, _maturity, _maturity_business_days, _tenor, _elapsed_days)
        else:
            raise ValueError("obs_type must be European or American")

        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            return value  # 终止日, spot是价格向量, 则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        else:
            return np.mean(value)

    @staticmethod
    def _preprocess_extra_returns(prod, period):
        """预处理可能存在的额外payoff：现金返还rebate、保底收益率floor_yield
        Args:
            prod: Product产品对象
            period: float 或 np.ndarray, 从起始到期权终止的时间（年）
                    对欧式障碍、美式敲入而言是整个存续期；对美式敲出而言，期权可能会因为敲出提前终止，此时period是起始日到敲出日的时间
        Returns:
            _floor_yield: float，处理后的保底收益绝对数值
            _rebate: float，处理后的现金返还绝对数值
        """
        if prod.floor_yield != 0:  # 保底收益率
            _floor_yield = prod.floor_yield * period * prod.s0
        else:
            _floor_yield = 0.0
        if prod.rebate != 0:  # 现金返还
            if prod.rebate_annual:  # 年化百分比
                _rebate = prod.rebate * period * prod.s0
            else:
                _rebate = prod.rebate  # 绝对数值
        else:
            _rebate = 0.0
        return _floor_yield, _rebate

    def _calc_european_barrier(self, prod, paths, _maturity, _tenor):
        """欧式障碍期权(到期观察)的蒙特卡洛模拟
        Args:
            prod: Product产品对象
            paths: np.ndarray，模拟价格路径
            _maturity: float，从估值日到终止日的自然日时间（年）
            _tenor: float, 从起始日到终止日的自然日时间（年）
        Returns: float，欧式障碍期权(到期观察)的估值结果
        """
        if prod.updown == UpDown.Up:
            knock_inout = (paths[-1] >= prod.barrier)
        elif prod.updown == UpDown.Down:
            knock_inout = (paths[-1] <= prod.barrier)
        else:
            raise ValueError("不支持的UpDown类型")

        _floor_yield, _rebate = self._preprocess_extra_returns(prod, _tenor)

        if prod.inout == InOut.In:
            payoff = np.ones(paths[-1].size) * _rebate
            payoff[knock_inout] = np.maximum(prod.callput.value * (paths[-1, knock_inout] - prod.strike),
                                             0) * prod.parti + _floor_yield
        elif prod.inout == InOut.Out:
            payoff = np.maximum(prod.callput.value * (paths[-1] - prod.strike), 0) * prod.parti + _floor_yield
            payoff[knock_inout] = _rebate
        else:
            raise ValueError("不支持的InOut类型")
        payoff = payoff * self.process.interest.disc_factor(_maturity)
        return payoff

    def _calc_american_barrier(self, prod, paths, _maturity, _maturity_business_days, _tenor, _elapsed_days):
        """美式障碍期权的蒙特卡洛模拟
        Args:
            prod: Product产品对象
            paths: np.ndarray，模拟价格路径
            _maturity: float，从估值日到终止日的自然日到期时间（年）
            _maturity_business_days: int，从估值日到终止日的交易日到期时间（天）
            _tenor: float, 从起始日到终止日的自然日存续时间（年）
            _elapsed_days: int, 从起始日到估值日的交易日时间（天）
        Returns: float，美式障碍期权的估值结果
        """
        if prod.discrete_obs_interval is None:  # 每日观察
            obs_points = np.arange(0, _maturity_business_days + 1, 1)
        else:  # 均匀离散观察
            dt_step = prod.discrete_obs_interval * prod.t_step_per_year
            obs_points = np.flip(np.round(np.arange(_maturity_business_days, 0, -dt_step)).astype(int))
            obs_points = np.concatenate((np.array([0]), obs_points))
            obs_points[obs_points > _maturity_business_days] = _maturity_business_days  # 防止闰年导致的下标越界

        if prod.updown == UpDown.Up:
            knock_inout = np.any(paths[obs_points] >= prod.barrier, axis=0)
        elif prod.updown == UpDown.Down:
            knock_inout = np.any(paths[obs_points] <= prod.barrier, axis=0)
        else:
            raise ValueError("不支持的UpDown类型")

        _floor_yield, _rebate = self._preprocess_extra_returns(prod, _tenor)
        if prod.inout == InOut.In:
            payoff = np.ones(paths[-1].size) * _rebate
            payoff[knock_inout] = np.maximum(prod.callput.value * (paths[-1, knock_inout] - prod.strike),
                                             0) * prod.parti + _floor_yield
            payoff = payoff * self.process.interest.disc_factor(_maturity)
            return payoff
        elif prod.inout == InOut.Out:
            payoff = (np.maximum(prod.callput.value * (paths[-1] - prod.strike), 0
                                 ) * prod.parti + _floor_yield) * self.process.interest.disc_factor(_maturity)
            if prod.payment_type == PaymentType.Expire:  # 敲出时，到期再支付现金返还
                payoff[knock_inout] = _rebate * self.process.interest.disc_factor(_maturity)
            elif prod.payment_type == PaymentType.Hit:  # 敲出时，立即支付现金返还
                step_index = np.tile(obs_points, (paths.shape[1], 1)).T.astype(int)
                if prod.updown == UpDown.Up:
                    hit_time = np.min(np.where(paths[obs_points] >= prod.barrier, step_index, np.inf), axis=0)
                elif prod.updown == UpDown.Down:
                    hit_time = np.min(np.where(paths[obs_points] <= prod.barrier, step_index, np.inf), axis=0)
                else:
                    raise ValueError("不支持的UpDown类型")
                if prod.rebate != 0:  # 现金返还
                    if prod.rebate_annual:  # 年化百分比
                        _rebate = prod.s0 * prod.rebate * (_elapsed_days + hit_time[hit_time != np.inf]
                                                           ) / prod.t_step_per_year
                    else:
                        _rebate = prod.rebate  # 绝对数值
                else:
                    _rebate = 0.0
                payoff[knock_inout] = _rebate * self.process.interest.disc_factor(
                    hit_time[hit_time != np.inf] / prod.t_step_per_year)
            else:
                raise ValueError("PaymentType must be Hit or Expire")
            return payoff
        else:
            raise ValueError("不支持的InOut类型")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
import numpy as np
from pricelib.common.utilities.enums import (CallPut, InOut, UpDown, BarrierType, PaymentType, EngineType, StatusType,
                                             ExerciseType)
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer, global_auto_update_status
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.analytic_engines import AnalyticBarrierEngine
from pricelib.pricing_engines.fdm_engines import FdmVanillaEngine
from pricelib.products.vanilla.vanilla_option import VanillaOption
from pricelib.products.barrier.barrier_option import BarrierOption


class SingleShark(BarrierOption):
    """单向鲨鱼鳍，含封底价格"""

    def __init__(self, strike, barrier, rebate, callput=CallPut.Call, updown=UpDown.Up, *, parti=1,
                 obs_type=ExerciseType.American, payment_type=None, rebate_annual=False, floor_yield=0.0,
                 maturity=None, start_date=None, end_date=None, discrete_obs_interval=None, status=StatusType.NoTouch,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 engine=None, stoch_process=None, s0=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """构造函数
        产品参数:
            strike: float, 执行价
            barrier: float, 障碍价
            rebate: float, 敲出现金返还金额，或未敲入现金返还金额，绝对数值，非百分比，非年化
            rebate_annual: bool, 现金返还是否为年化，默认为False，即绝对数值，非年化，非百分比；若为True，则为百分比年化收益率
            callput: 看涨看跌类型，CallPut枚举类，Call看涨/Put看跌
            inout: 敲入敲出类型，InOut枚举类，In敲入/Out敲出
            updown: 向上向下类型，UpDown枚举类，Up向上/Down向下
            payment_type: 现金返还支付类型，PaymentType枚举类，敲入默认为到期支付Expire；敲出默认为立即支付Hit
            parti: float, 香草期权的参与率，默认为1
            floor_yield: float, 封底收益(保底收益)，当未敲出情况下，约定的最低收益率，默认为0，绝对数值，非百分比，非年化。
                          即未敲出收益支付 = 封底收益 + 参与率 * max[(期末标的价-行权价格), 0]
            status: 估值日前的敲入敲出状态，StatusType枚举类，默认为NoTouch未敲入/未敲出，UpTouch为向上敲入/敲出，DownTouch为向下敲入/敲出
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
            discrete_obs_interval: 观察时间间隔. 若为连续观察或到期观察，None；若为均匀离散观察，为年化的观察时间间隔(float)。
        可选参数:
            s0: float，标的期初价格，只有当 rebate_annual = True，即现金返还为年化收益率时；
                                  或者当 floor_yield != 0，即存在非零的保底收益率时，
                       需要输入标的期初价格，将百分比的收益率转化成绝对数值
            engine: 定价引擎，PricingEngine类，以下几种引擎均支持向上/向下、敲入/敲出、看涨/看跌8种单边障碍期权，支持敲出/未敲入现金返还
                            对于美式障碍，解析解和PDE引擎支持连续观察，其余引擎只支持离散观察(默认为每日观察)
                    解析解: AnalyticBarrierEngine 支持欧式障碍（到期观察）和美式障碍（支持连续观察/离散观察）
                                        欧式障碍：到期观察，支持保底收益率，支持支持现金返还为绝对数值/年化百分比
                                        美式障碍：敲入现金返还为到期支付；敲出现金返还为立即支付。现金返还为固定绝对数值，不支持年化的现金返还。
                                                不支持同时输入非零的现金返还和不等于1的参与率，二者只有其一时支持。不支持保底收益率
                    蒙特卡洛: MCBarrierEngine 支持欧式障碍（到期观察）和美式障碍（仅支持离散观察）
                                        欧式障碍：到期观察，支持保底收益率，支持支持现金返还为绝对数值/年化百分比
                                        美式障碍 ：敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
                                                支持现金返还为绝对数值/年化百分比，支持保底收益率
                    PDE: FdmBarrierEngine  支持欧式障碍（到期观察）和美式障碍（支持连续观察/离散观察）
                                        欧式障碍：到期观察，支持保底收益率，支持支持现金返还为绝对数值/年化百分比
                                        美式障碍 ：敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
                                                支持现金返回为绝对数值/年化百分比，支持保底收益率
                    积分法: QuadBarrierEngine  仅支持离散观察的美式障碍，todo: 支持欧式障碍、年化的现金返还、保底收益率
                                            敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
                    二叉树: BiTreeBarrierEngine  仅支持离散观察的美式障碍，todo: 支持欧式障碍、年化的现金返还、保底收益率
                                            敲入现金返还为到期支付；敲出现金返还 支持 立即支付/到期支付
            engine_method: str, 定价引擎种类。若未输入engine，可以用engine_method快捷生成指定定价引擎。可选Analytic、MC/MonteCarlo、PDE/FDM、QUAD、TREE。
            若未提供引擎engine、也未指定引擎种类engine_method的情况下，如果
            若未提供引擎的情况下，提供了 随机过程 或 标的价格、无风险利率、分红/融券率、波动率，
            则默认使用解析解定价引擎 AnalyticBarrierEngine - Merton(1973), Reiner & Rubinstein(1991a)
                                支持 Broadie, Glasserman和Kou(1995)均匀离散观察调整
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        self._inout = InOut.Out
        # 连续观察/到期观察，discrete_obs_interval=None; 均匀离散观察=观察时间间隔; 每日观察自动转为1/243
        discrete_obs_interval = 1 / 243 if discrete_obs_interval == "每日观察" else discrete_obs_interval
        super().__init__(inout=InOut.Out, discrete_obs_interval=discrete_obs_interval, obs_type=obs_type,
                         strike=strike, barrier=barrier, callput=callput, updown=updown, parti=parti,
                         rebate=rebate, rebate_annual=rebate_annual, floor_yield=floor_yield,
                         payment_type=payment_type, status=status, maturity=maturity, start_date=start_date,
                         end_date=end_date, trade_calendar=trade_calendar, annual_days=annual_days,
                         t_step_per_year=t_step_per_year, engine=engine, stoch_process=stoch_process,
                         s0=s0, s=s, r=r, q=q, vol=vol, engine_method=engine_method)

    def __repr__(self):
        """返回期权的描述"""
        if self.obs_type == ExerciseType.European:
            obs_type = "到期观察"
        elif self.obs_type == ExerciseType.American:
            if self.discrete_obs_interval is None:
                obs_type = "连续观察"
            elif self.discrete_obs_interval == 1 / 243:
                obs_type = "每日观察"
            else:
                obs_type = "离散观察"
        else:
            raise ValueError("obs_type should be European or American")
        return f"{obs_type}{self._callput.name}单向鲨鱼鳍期权"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        if self.status != StatusType.NoTouch:  # 如果已经触碰障碍，则敲出价值=现金返还的现值
            calculate_date = global_evaluation_date() if t is None else t
            if self.payment_type == PaymentType.Hit:  # 立即支付
                if self.rebate_annual:  # 年化百分比
                    elapsed_time = (calculate_date - self.start_date).days / self.annual_days.value
                    _rebate = [r * self.s0 * elapsed_time for r in self.rebate]
                else:  # 绝对数值
                    _rebate = self.rebate
                return _rebate
            elif self.payment_type == PaymentType.Expire:  # 到期支付
                if self.rebate_annual:  # 年化百分比
                    tenor = (self.end_date - self.start_date).days / self.annual_days.value  # 从起始日到终止日的自然日时间（年）
                    _rebate = [r * self.s0 * tenor for r in self.rebate]
                else:  # 绝对数值
                    _rebate = self.rebate
                _maturity = (self.end_date - calculate_date).days / self.annual_days.value
                return self.rebate * self.engine.process.interest.disc_factor(_maturity)
            else:
                raise ValueError(f"未知的支付方式{self.payment_type}")
        else:  # 尚未触碰障碍，正常定价
            if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边障碍，所以需要把barrier填入bound
                if self.updown == UpDown.Up:
                    bound = (None, self.barrier)
                else:
                    bound = (self.barrier, None)
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=bound,
                                                       rebate=(self.rebate, self.rebate))
            else:
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



# !/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
import numpy as np
from pricelib.common.utilities.enums import InOut, ExerciseType, EngineType, PaymentType, StatusType
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer, global_auto_update_status
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.fdm_engines import FdmDoubleSharkEngine
from pricelib.pricing_engines.mc_engines import MCDoubleSharkEngine
from pricelib.pricing_engines.integral_engines import QuadDoubleSharkEngine


class DoubleShark(OptionBase, Observer):
    """双向鲨鱼鳍结构
    双向鲨鱼鳍结构属于敲出期权，由两个高低障碍价相同的双边看涨、双边看跌敲出障碍期权组合而成，且put的执行价小于等于call的执行价"""
    inout = InOut.Out  # 双鲨结构属于敲出期权

    def __init__(self, obs_type=ExerciseType.American, payment_type=PaymentType.Expire,
                 strike_lower=95., strike_upper=105., barrier_lower=90., barrier_upper=110.,
                 parti_lower=1., parti_upper=1., rebate_lower=0., rebate_upper=0., rebate_annual=False,
                 floor_yield=0., status=StatusType.NoTouch, window=(None, None),
                 maturity=None, start_date=None, end_date=None, discrete_obs_interval=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 engine=None, stoch_process=None, s0=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """构造函数
        产品参数:
            obs_type: 障碍观察类型，ExerciseType枚举类，美式观察American/欧式到期观察European
            payment_type: 支付类型，PaymentType枚举类，默认为到期支付Expire
            strike: (strike_lower, strike_upper), 低行权价/高行权价，一般低行权价是put的执行价，高行权价是call的执行价
            bound: (barrier_lower, barrier_upper), 低障碍价/高障碍价，障碍价格绝对数值。如果是默认值None，即为无穷大或无穷小inf。
            parti: (parti_lower, parti_upper), 香草期权的参与率，默认为1
            rebate: (rebate_lower, rebate_upper), 未敲出补偿的绝对数值，非年化
                                                  元组第一位是低障碍价下方的补偿，第二位是高障碍价上方的补偿，
            rebate_annual: bool, 现金返还是否为年化，默认为False，即绝对数值，非年化，非百分比；若为True，则为百分比年化收益率
            floor_yield: float, 封底收益(保底收益)，默认为0，未敲出情况下，约定的最低收益率，即 max(|S-K|,0) 额外的固定收益
            status: 敲出状态，StatusType枚举类，默认为NoTouch未敲出，UpTouch为向上敲出，DownTouch为向下敲出
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            window: (start_date, end_date)，障碍观察时间窗口，距离估值日的年化期限。 仅PDE定价引擎支持该参数。
                    如果是默认值None，说明观察区间是整个存续期: 估值日-终止日
            discrete_obs_interval: 观察时间间隔. 若为连续观察，None；若为均匀离散观察，为年化的观察时间间隔
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            s0: float，标的期初价格，只有当 rebate_annual = True，即现金返还为年化收益率时；
                      或者当 floor_yield != 0，即存在非零的保底收益率时，需要输入标的期初价格，将百分比的收益率转化成绝对数值
            engine: 定价引擎，PricingEngine类
                    解析解: AnalyticDoubleSharkEngine 分为两种，只支持美式观察(整个有效期观察)；支持连续观察/离散观察(默认为每日观察)；
                                                    Ikeda and Kunitomo(1992) 仅当行权价在障碍范围内时，公式才是成立的
                                                    Haug(1998) 仅适用于标的持有成本为0的期权(期货期权)
                                    两者的现金返还部分都使用美式双接触闭式解Hui(1996)，因此高低敲出现金返还也必须相等，现金返还只支持到期支付
                    PDE: FdmDoubleSharkEngine  只支持美式观察(整个有效期观察)；支持连续观察/离散观察(默认为每日观察)；
                                               支持现金返还；敲出现金返还支持 立即支付/到期支付
                    蒙特卡洛: MCDoubleSharkEngine  支持欧式观察(仅到期观察)/美式观察(整个有效期观察)；只支持离散观察(默认为每日观察)；
                                                 支持现金返还；敲出现金返还支持 立即支付/到期支付
                    积分法: QuadDoubleSharkEngine  只支持美式观察(整个有效期观察)；只支持离散观察(默认为每日观察)；支持现金返还到期支付；
            engine_method: str, 定价引擎种类。若未输入engine，可以用engine_method快捷生成指定定价引擎。可选Analytic、MC/MonteCarlo、PDE/FDM、QUAD。
            若未提供引擎engine、也未指定引擎种类engine_method的情况下，如果提供了 随机过程 或 标的价格、无风险利率、分红/融券率、波动率，
            则默认使用PDE定价引擎 FdmDoubleSharkEngine
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if self.trade_calendar.isholiday(self.end_date):  # 如果到期日不是交易日，向后顺延一个交易日
            self.end_date = self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1))
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        assert strike_lower <= strike_upper, "ValueError: 双鲨结构的低行权价必须小于等于高行权价"
        self.strike = (strike_lower, strike_upper)  # (低行权价, 高行权价)
        assert barrier_lower < barrier_upper, "ValueError: 双鲨结构的低障碍价必须小于高障碍价"
        self.bound = (barrier_lower, barrier_upper)  # (低障碍价, 高障碍价)
        self.parti = (parti_lower, parti_upper)  # (低参与率，高参与率)
        self.rebate = (rebate_lower, rebate_upper)  # 敲出现金返还: (下边界payoff, 上边界payoff)
        self.rebate_annual = rebate_annual  # 现金返还是否为年化
        self.window = window  # (观察窗口起始时间, 观察窗口结束时间)
        self.floor_yield = floor_yield  # 封底收益
        self.obs_type = obs_type
        self.s0 = s0
        self.payment_type = payment_type
        self.discrete_obs_interval = discrete_obs_interval  # 连续观察=None；均匀离散观察=观察时间间隔
        if self.obs_type == ExerciseType.European:
            err_msg = ""
            if self.payment_type != PaymentType.Expire:
                err_msg += "ValueError: 到期观察双鲨的现金返还方式PaymentType一定是到期支付Expire.\n"
            if self.discrete_obs_interval is not None:
                err_msg += "ValueError: 到期观察的双向鲨鱼鳍，观察时间间隔只能为None.\n"
            if err_msg != "":
                raise ValueError(err_msg)
        self.status = status
        self.stop_date = self.end_date  # 停止日期，默认等于到期日。如果发生敲出并立即支付现金返还，停止日期更新为敲出日期
        if engine is not None:
            self.set_pricing_engine(engine)
        else:
            if engine_method is None:
                default_engine = FdmDoubleSharkEngine  # 默认PDE-FDM定价引擎
            elif engine_method == "Analytic":
                from pricelib.pricing_engines.analytic_engines import AnalyticDoubleSharkEngine
                default_engine = AnalyticDoubleSharkEngine
            elif engine_method in ["PDE", "FDM"]:
                default_engine = FdmDoubleSharkEngine
            elif engine_method in ["MC", "MonteCarlo"]:
                default_engine = MCDoubleSharkEngine
            elif engine_method == "QUAD":
                default_engine = QuadDoubleSharkEngine
            else:
                raise NotImplementedError(
                    f"不支持的engine_method{engine_method}, {self.__repr__()}仅支持Analytic/MC/PDE/QUAD")
            if stoch_process is not None:
                self.set_pricing_engine(default_engine(stoch_process))
            elif s is not None and r is not None and q is not None and vol is not None:
                self.set_pricing_engine(default_engine(s=s, r=r, q=q, vol=vol))

    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经触碰障碍"""  # TODO: 目前update仅支持连续观察，或者每日观察
        if global_evaluation_date() > self.stop_date:  # 如果估值日已经过了终止日，维持原来的状态
            return
        if global_auto_update_status() and observable == self.engine.process.spot:
            if self.obs_type == ExerciseType.American:
                # TODO: 目前美式观察双边障碍期权的update仅支持连续观察，或者每日观察（因为每周、每月观察日不确定，需要额外输入参数）
                if observable.data >= self.bound[1]:
                    self.status = StatusType.UpTouch
                    if self.inout == InOut.Out and self.payment_type == PaymentType.Hit:
                        self.stop_date = global_evaluation_date()
                        logging.info(f"{self}【已向上触碰】，终止日为 {self.stop_date}")
                elif observable.data <= self.bound[0]:
                    self.status = StatusType.DownTouch
                    if self.inout == InOut.Out and self.payment_type == PaymentType.Hit:
                        self.stop_date = global_evaluation_date()
                        logging.info(f"{self}【已向下触碰】，终止日为 {self.stop_date}")
            elif self.obs_type == ExerciseType.European:  # 欧式到期观察
                if global_evaluation_date() == self.end_date:  # 如果估值日是到期日
                    if observable.data >= self.bound[1]:
                        self.status = StatusType.UpTouch
                    elif observable.data <= self.bound[0]:
                        self.status = StatusType.DownTouch

    def __repr__(self):
        """返回期权的描述"""
        if self.obs_type == ExerciseType.European:
            obs_type = "到期观察"
        elif self.obs_type == ExerciseType.American:
            if self.discrete_obs_interval is None:
                obs_type = "连续观察"
            elif self.discrete_obs_interval == 1 / 243:
                obs_type = "每日观察"
            else:
                obs_type = "离散观察"
        else:
            raise ValueError("obs_type should be European or American")
        return f"{obs_type}{self.payment_type.value}双向鲨鱼鳍期权"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        assert (self.strike[0] <= self.strike[1]) and (
                self.bound[0] < self.bound[1]), "Error: 双鲨结构输入的行权价和障碍价必须是递增的，不能是None"
        if self.status != StatusType.NoTouch:  # 如果已经触碰障碍，则敲出价值=现金返还的现值
            calculate_date = global_evaluation_date() if t is None else t
            tenor = (self.end_date - self.start_date).days / self.annual_days.value  # 从起始日到终止日的自然日时间（年）
            _maturity = (self.end_date - calculate_date).days / self.annual_days.value
            # 已经发生敲出，支付现金返还rebate
            if self.payment_type == PaymentType.Hit:  # 立即支付
                if self.rebate_annual:  # 年化百分比
                    elapsed_time = (calculate_date - self.start_date).days / self.annual_days.value
                    _rebate = [r * self.s0 * elapsed_time for r in self.rebate]
                else:  # 绝对数值
                    _rebate = self.rebate
                if self.status == StatusType.UpTouch:  # 向上触碰
                    return _rebate[1]
                else:  # 向下触碰
                    return _rebate[0]
            elif self.payment_type == PaymentType.Expire:  # 到期支付
                if self.rebate_annual:  # 年化百分比
                    _rebate = [r * self.s0 * tenor for r in self.rebate]
                else:  # 绝对数值
                    _rebate = self.rebate
                if self.status == StatusType.UpTouch:  # 向上触碰
                    return _rebate[1] * self.engine.process.interest.disc_factor(_maturity)
                else:  # 向下触碰
                    return _rebate[0] * self.engine.process.interest.disc_factor(_maturity)
            else:
                raise ValueError(f"未知的支付方式{self.payment_type}")
        else:  # 尚未触碰障碍，正常定价
            if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边障碍
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=self.bound,
                                                       rebate=self.rebate, window=self.window)
            else:
                price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import pytest
import datetime
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from ..conftest import init_bsm_process
from pricelib import *
from examples.autocallable import (phoenix_demo, standard_snowball_demo, earlyprofit_snowball_demo,
                                   butterfly_snowball_demo, otm_snowball_demo, floored_snowball_demo,
                                   stepdown_snowball_demo, bothdown_snowball_demo, parachute_snowball_demo,
                                   snowball_plus_demo, paris_snowball_demo)


class TestAutocallNote:
    """测试Autocall Note(二元小雪球)"""

    @staticmethod
    def run(evaluation_date: datetime.date, spot: float):
        process = init_bsm_process(evaluation_date, s=spot, r=0.03, q=0.05, vol=0.2)
        # 3. 定价引擎，包括蒙特卡洛模拟、有限差分、数值积分
        mc_engine = MCAutoCallEngine(process, n_path=100000, rands_method=RandsMethod.Pseudorandom,
                                     antithetic_variate=True, ld_method=LdMethod.Halton, seed=0)
        quad_engine = QuadAutoCallEngine(process, quad_method=QuadMethod.Simpson, n_points=1301)
        pde_engine = FdmSnowBallEngine(process, s_step=800, n_smax=2, fdm_theta=1)
        # 4. 定义产品：Autocall Note(二元小雪球)
        results = pd.DataFrame()
        for callput in CallPut:
            option = AutoCall(s0=100, maturity=2, start_date=datetime.date(2022, 1, 5), lock_term=3,
                              barrier_out=103, coupon_out=0.045, coupon_div=0.02, callput=callput, obs_dates=None,
                              pay_dates=None, margin_lvl=1, trade_calendar=CN_CALENDAR, t_step_per_year=243,
                              engine=None)
            # 5.为产品设置定价引擎
            option.set_pricing_engine(mc_engine)
            price_mc = option.pv_and_greeks()

            option.set_pricing_engine(pde_engine)
            price_pde = option.pv_and_greeks()

            option.set_pricing_engine(quad_engine)
            price_quad = option.pv_and_greeks()

            df = pd.DataFrame([price_mc, price_pde, price_quad], index=['MonteCarlo', 'PDE', 'Quadrature'])
            df.insert(0, "期权类型", str(option))
            results = pd.concat([results, df], axis=0)

        results = results.reset_index()
        return results

    def test_first_day(self):
        """测试期初估值"""
        res_df = self.run(evaluation_date=datetime.date(2022, 1, 5), spot=100)
        expected_data = [{'index': 'MonteCarlo', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 99.93100555332597,
                          'delta': 0.04560718511393702, 'gamma': -0.0017764579370975753, 'vega': 0.006866722182664375,
                          'theta': 0.009587446233922492, 'rho': -0.8369559555542452},
                         {'index': 'PDE', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 99.96033974577038,
                          'delta': 0.04474355948097042, 'gamma': -0.002709507283910284, 'vega': 0.0057905353147731375,
                          'theta': 0.014941219562047081, 'rho': -0.8248703827510013},
                         {'index': 'Quadrature', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 99.96258810435373,
                          'delta': 0.03928482280267076, 'gamma': -0.001757213484637532, 'vega': -0.00032291027079622836,
                          'theta': 0.01507381313852818, 'rho': -0.8221706577589032},
                         {'index': 'MonteCarlo', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': 100.33861697900203,
                          'delta': -0.006719327618149862, 'gamma': 0.0009661924013215639,
                          'vega': -0.0011945570559532825, 'theta': 0.010620002727208089, 'rho': -0.5299448769754207},
                         {'index': 'PDE', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': np.nan, 'delta': np.nan,
                          'gamma': np.nan, 'vega': np.nan, 'theta': np.nan, 'rho': np.nan},
                         {'index': 'Quadrature', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': 100.36871158225928,
                          'delta': -0.006662976297697298, 'gamma': -0.001045828316065922, 'vega': 9.90501560949042e-05,
                          'theta': 0.013353193558771181, 'rho': -0.5263612203634693}]
        expected_df = pd.DataFrame(expected_data)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_last_day_1(self):
        """测试到期日, 敲出价上方"""
        res_df = self.run(evaluation_date=datetime.date(2024, 1, 5), spot=105)
        expected_data = [
            {'index': 'MonteCarlo', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 109.00000000000001, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'PDE', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 109.00000000000001, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'Quadrature', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 109.00000000000001, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'MonteCarlo', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': 104.0, 'delta': 0.0, 'gamma': 0.0,
             'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'PDE', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': np.nan, 'delta': np.nan, 'gamma': np.nan,
             'vega': np.nan, 'theta': np.nan, 'rho': np.nan},
            {'index': 'Quadrature', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': 104.0, 'delta': 0.0, 'gamma': 0.0,
             'vega': 0.0, 'theta': np.nan, 'rho': 0.0}]
        expected_df = pd.DataFrame(expected_data)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_last_day_2(self):
        """测试到期日, 敲出价下方"""
        res_df = self.run(evaluation_date=datetime.date(2024, 1, 5), spot=95)
        expected_data = [
            {'index': 'MonteCarlo', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 104.0, 'delta': 0.0, 'gamma': 0.0,
             'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'PDE', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 104.0, 'delta': 0.0, 'gamma': 0.0,
             'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'Quadrature', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 104.0, 'delta': 0.0, 'gamma': 0.0,
             'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'MonteCarlo', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': 109.00000000000001, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'PDE', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': np.nan, 'delta': np.nan, 'gamma': np.nan,
             'vega': np.nan, 'theta': np.nan, 'rho': np.nan},
            {'index': 'Quadrature', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': 109.00000000000001, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0}]
        expected_df = pd.DataFrame(expected_data)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_penultimate_day(self):
        """测试倒数第二天"""
        res_df = self.run(evaluation_date=datetime.date(2024, 1, 4), spot=102)
        expected_data = [{'index': 'MonteCarlo', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 105.09101202759277,
                          'delta': 1.0911603119050381, 'gamma': 0.5841984253932281, 'vega': 0.05534545087188292,
                          'theta': -1.0910120275927682, 'rho': 0.002070291195806817},
                         {'index': 'PDE', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 104.93900183414304,
                          'delta': 1.2475814729859098, 'gamma': 1.2166670865011868, 'vega': 0.04448723733024451,
                          'theta': -0.9390018341430704, 'rho': -8.250464362902221e-05},
                         {'index': 'Quadrature', '期权类型': 'AutoCall Note(二元小雪球)', 'pv': 105.07928482696474,
                          'delta': 1.0715455441597908, 'gamma': 0.575234196548749, 'vega': 0.045874167127550436,
                          'theta': -1.079284826964738, 'rho': 0.0004103781644744231},
                         {'index': 'MonteCarlo', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': 107.89148184253199,
                          'delta': -1.091160311905045, 'gamma': -0.5841984253932144, 'vega': -0.05534545087188292,
                          'theta': 1.1085181574680263, 'rho': -0.007905348081493457},
                         {'index': 'PDE', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': np.nan, 'delta': np.nan,
                          'gamma': np.nan, 'vega': np.nan, 'theta': np.nan, 'rho': np.nan},
                         {'index': 'Quadrature', '期权类型': 'AutoPut Note(二元小雪球)', 'pv': 107.89442050012653,
                          'delta': -1.0715455441598047, 'gamma': -0.5752341965502787, 'vega': -0.04587416700508129,
                          'theta': 1.1055794998734854, 'rho': -0.009174547843997516}]
        expected_df = pd.DataFrame(expected_data)
        assert_frame_equal(res_df, expected_df, check_dtype=True)


class TestFCN:
    """测试FCN"""

    @staticmethod
    def run(evaluation_date: datetime.date, spot: float):
        process = init_bsm_process(evaluation_date, s=spot, r=0.02, q=0.04, vol=0.16)
        # 3. 定价引擎，包括蒙特卡洛模拟、有限差分、数值积分
        mc_engine = MCPhoenixEngine(process, n_path=100000, rands_method=RandsMethod.Pseudorandom,
                                    antithetic_variate=True, ld_method=LdMethod.Sobol, seed=0)
        quad_engine = QuadFCNEngine(process, quad_method=QuadMethod.Simpson, n_points=1301)
        pde_engine = FdmPhoenixEngine(process, s_step=800, n_smax=2, fdm_theta=1)

        # 4. 定义产品：FCN(Fixed Coupon Note固定派息票据)
        option = FCN(maturity=2, s0=100, start_date=datetime.date(2022, 1, 5), trade_calendar=CN_CALENDAR,
                     barrier_out=103, barrier_in=80, strike_upper=None, coupon=0.00314, lock_term=3,
                     engine=None, status=StatusType.NoTouch, t_step_per_year=243)
        # 5.为产品设置定价引擎
        option.set_pricing_engine(mc_engine)
        price_mc = option.pv_and_greeks()

        option.set_pricing_engine(pde_engine)
        price_pde = option.pv_and_greeks()

        option.set_pricing_engine(quad_engine)
        price_quad = option.pv_and_greeks()

        df = pd.DataFrame([price_mc, price_pde, price_quad], index=['MonteCarlo', 'PDE', 'Quadrature'])
        df.insert(0, "期权类型", str(option))
        df = df.reset_index()
        return df

    def test_first_day(self):
        """测试期初估值"""
        res_df = self.run(evaluation_date=datetime.date(2022, 1, 5), spot=100)
        expected = [{'index': 'MonteCarlo', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.05171787086645,
                     'delta': 0.07635024390848599, 'gamma': -0.015165972870889277, 'vega': -0.2387282542682101,
                     'theta': 0.012701412348405938, 'rho': -0.7136221312761393},
                    {'index': 'PDE', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.00675403778338,
                     'delta': 0.08314976924964412, 'gamma': -0.011802049851411311, 'vega': -0.24197337781235717,
                     'theta': 0.015131337693560454, 'rho': -0.699107384765469},
                    {'index': 'Quadrature', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.00714124444022,
                     'delta': 0.08163928634644435, 'gamma': -0.005477769623979611, 'vega': -0.2443748254342495,
                     'theta': 0.015040570251045438, 'rho': -0.7035455452124779}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_knockout_day(self):
        """测试敲出观察日，敲出线上方"""
        res_df = self.run(evaluation_date=datetime.date(2023, 1, 5), spot=104)
        expected = [{'index': 'MonteCarlo', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.314,
                     'delta': -0.16258168039846246, 'gamma': 0.31265707768935086, 'vega': 0.0,
                     'theta': -0.0036999895239517855, 'rho': 0.0},
                    {'index': 'PDE', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.314,
                     'delta': -0.019744791561714638, 'gamma': 0.03796955361753901, 'vega': 0.0,
                     'theta': -0.007857451128046478, 'rho': 0.0},
                    {'index': 'Quadrature', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.314,
                     'delta': -0.1590626081067329, 'gamma': 0.3058896309744863, 'vega': 0.0,
                     'theta': 0.0027993324027875133, 'rho': 0.0}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_last_day_1(self):
        """测试到期日，未敲入未敲出"""
        res_df = self.run(evaluation_date=datetime.date(2024, 1, 5), spot=100)
        expected = [
            {'index': 'MonteCarlo', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.314, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'PDE', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.314, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'Quadrature', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.314, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_last_day_2(self):
        """测试到期日，敲入"""
        res_df = self.run(evaluation_date=datetime.date(2024, 1, 5), spot=79)
        expected = [{'index': 'MonteCarlo', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 99.314,
                     'delta': 0.9999999999999449, 'gamma': 2.277015656978369e-14, 'vega': 0.0, 'theta': np.nan,
                     'rho': 0.0},
                    {'index': 'PDE', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 99.314,
                     'delta': 0.9999999999999989,
                     'gamma': 2.277015656978369e-14, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
                    {'index': 'Quadrature', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 99.314,
                     'delta': 1.0000000000000078, 'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_penultimate_day(self):
        """测试倒数第二天"""
        res_df = self.run(evaluation_date=datetime.date(2024, 1, 4), spot=82)
        expected = [{'index': 'MonteCarlo', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.30590069537448,
                     'delta': 0.018144003437236284, 'gamma': -0.0368077052172507, 'vega': -0.0015630135122108868,
                     'theta': 0.008099304625517334, 'rho': -0.0027166507625224767},
                    {'index': 'PDE', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.29562497052231,
                     'delta': 0.022477502654348103, 'gamma': -0.032539236461599265, 'vega': -0.002955125070457143,
                     'theta': 0.018375029477681437, 'rho': -0.003972152545600238},
                    {'index': 'Quadrature', '期权类型': 'FCN(Fixed Coupon Note固定派息票据)', 'pv': 100.30346702051233,
                     'delta': 0.01769893847376848, 'gamma': -0.03663982148852997, 'vega': -0.0013905131321791941,
                     'theta': 0.010532979487663852, 'rho': -0.0041003354883173415}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)


class TestDCN:
    """测试DCN"""

    @staticmethod
    def run(evaluation_date: datetime.date, spot: float):
        process = init_bsm_process(evaluation_date, s=spot, r=0.02, q=0.04, vol=0.16)
        # 3. 定价引擎，包括蒙特卡洛模拟、有限差分、数值积分
        mc_engine = MCPhoenixEngine(process, n_path=100000, rands_method=RandsMethod.Pseudorandom,
                                    antithetic_variate=True, ld_method=LdMethod.Sobol, seed=0)
        quad_engine = QuadFCNEngine(process, quad_method=QuadMethod.Simpson, n_points=1301)
        pde_engine = FdmPhoenixEngine(process, s_step=800, n_smax=2, fdm_theta=1)

        # 4. 定义产品：DCN(Digital Coupon Note 二元派息票据)
        option = DCN(maturity=2, s0=100, start_date=datetime.date(2022, 1, 5), trade_calendar=CN_CALENDAR,
                     barrier_out=103, barrier_in=80, strike_upper=None, barrier_yield=80, coupon=0.00387, lock_term=3,
                     engine=None, status=StatusType.NoTouch, t_step_per_year=243)
        # 5.为产品设置定价引擎
        option.set_pricing_engine(mc_engine)
        price_mc = option.pv_and_greeks()

        option.set_pricing_engine(pde_engine)
        price_pde = option.pv_and_greeks()

        option.set_pricing_engine(quad_engine)
        price_quad = option.pv_and_greeks()

        df = pd.DataFrame([price_mc, price_pde, price_quad], index=['MonteCarlo', 'PDE', 'Quadrature'])
        df.insert(0, "期权类型", str(option))
        df = df.reset_index()
        return df

    def test_first_day(self):
        """测试期初估值"""
        res_df = self.run(evaluation_date=datetime.date(2022, 1, 5), spot=100)
        expected = [
            {'index': 'MonteCarlo', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.03897337641266,
             'delta': 0.12449349817369182, 'gamma': -0.023520650192580206, 'vega': -0.33071415418423555,
             'theta': 0.016330219417440617, 'rho': -0.6463332114506528},
            {'index': 'PDE', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.00429339431805,
             'delta': 0.1290889351727742, 'gamma': -0.019322770749539586, 'vega': -0.3328581754794442,
             'theta': 0.01946855195340902, 'rho': -0.6338018244863406},
            {'index': 'Quadrature', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.00228931538513,
             'delta': 0.13155965751516874, 'gamma': -0.02444031589277529, 'vega': -0.34232309973504016,
             'theta': 0.017786502542534777, 'rho': -0.6377361521656724}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_knockout_day_1(self):
        """测试敲出观察日，敲出线上方"""
        res_df = self.run(evaluation_date=datetime.date(2023, 1, 5), spot=105)
        expected = [
            {'index': 'MonteCarlo', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.387, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': 0.047595868779879424, 'rho': 0.0},
            {'index': 'PDE', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.38700000000001,
             'delta': -0.0002048632423709694, 'gamma': 0.0003902042532168839, 'vega': -1.4210854715202004e-14,
             'theta': 0.04883308078795778, 'rho': -2.842170943040401e-14},
            {'index': 'Quadrature', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.387, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': 0.061642633744000364, 'rho': 0.0}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_knockout_day_2(self):
        """测试敲出观察日，派息线上方、敲出线下方"""
        res_df = self.run(evaluation_date=datetime.date(2023, 1, 5), spot=90)
        expected = [
            {'index': 'MonteCarlo', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 99.87701603274787,
             'delta': 0.28609273949720027, 'gamma': -0.04228534469885087, 'vega': -0.3543265633827133,
             'theta': -0.36412160892682266, 'rho': -0.6243317922675118},
            {'index': 'PDE', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 99.76409482753948,
             'delta': 0.2899013976410537, 'gamma': -0.03844849802346168, 'vega': -0.36342304314723606,
             'theta': -0.3602739223918121, 'rho': -0.6169204411645381},
            {'index': 'Quadrature', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 99.75709566815144,
             'delta': 0.29114601692649145, 'gamma': -0.03763138514286106, 'vega': -0.37128410944430357,
             'theta': -0.3634817340838481, 'rho': -0.6181149756510536}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_knockout_day_3(self):
        """测试敲出观察日，派息线下方"""
        res_df = self.run(evaluation_date=datetime.date(2023, 1, 5), spot=75)
        expected = [
            {'index': 'MonteCarlo', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 90.52117091053566,
             'delta': 0.814720829067331, 'gamma': -0.022042791467457492, 'vega': -0.2069070484346014,
             'theta': 0.011878827572274986, 'rho': -0.35430189473626683},
            {'index': 'PDE', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 90.3889346797555,
             'delta': 0.8102488972281208, 'gamma': -0.008781525082415302, 'vega': -0.21707380306482094,
             'theta': 0.015030699199257924, 'rho': -0.3562835280030896},
            {'index': 'Quadrature', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 90.74952510342814,
             'delta': 0.8056141448776847, 'gamma': -0.007238400857608617, 'vega': -0.20251411243781092,
             'theta': -0.37011927564746827, 'rho': -0.357011880397053}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_last_day_1(self):
        """测试到期日，未敲入未敲出"""
        res_df = self.run(evaluation_date=datetime.date(2024, 1, 5), spot=100)
        expected = [
            {'index': 'MonteCarlo', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.387, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'PDE', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.387, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
            {'index': 'Quadrature', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.387, 'delta': 0.0,
             'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0}]
        expected_df = pd.DataFrame(expected)
        # assert_frame_equal(res_df, expected_df, check_dtype=True)
        print(res_df.to_dict('records'))  # 将DataFrame转换为字典列表
        pd.set_option('display.max_columns', None)  # 显示所有列
        print(res_df)

    def test_last_day_2(self):
        """测试到期日，敲入"""
        res_df = self.run(evaluation_date=datetime.date(2024, 1, 5), spot=79)
        expected = [{'index': 'MonteCarlo', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 99.0,
                     'delta': 0.9999999999999539, 'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
                    {'index': 'PDE', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 99.0,
                     'delta': 1.0000000000000078, 'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0},
                    {'index': 'Quadrature', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 99.0,
                     'delta': 1.0000000000000078, 'gamma': 0.0, 'vega': 0.0, 'theta': np.nan, 'rho': 0.0}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)

    def test_penultimate_day(self):
        """测试倒数第二天"""
        res_df = self.run(evaluation_date=datetime.date(2024, 1, 4), spot=82)
        expected = [
            {'index': 'MonteCarlo', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.37526296459806,
             'delta': 0.03658920612816049, 'gamma': -0.07150572954305838, 'vega': -0.0030064444179345173,
             'theta': 0.011737035401935714, 'rho': -0.0026605058436786067},
            {'index': 'PDE', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.36030176776694,
             'delta': 0.04096100061087157, 'gamma': -0.05930565090457058, 'vega': -0.004634434440873747,
             'theta': 0.026698232233030694, 'rho': -0.003876420038238848},
            {'index': 'Quadrature', '期权类型': 'DCN(Digital Coupon Note 二元派息票据)', 'pv': 100.37322979852743,
             'delta': 0.03637979787029794, 'gamma': -0.07302746997014209, 'vega': -0.0027675708252274944,
             'theta': 0.013770201472567578, 'rho': -0.004068063933814869}]
        expected_df = pd.DataFrame(expected)
        assert_frame_equal(res_df, expected_df, check_dtype=True)
        print(res_df.to_dict('records'))  # 将DataFrame转换为字典列表
        pd.set_option('display.max_columns', None)  # 显示所有列
        print(res_df)


def test_phoenix_demo():
    expected_data = {'结构': 'Phoenix 凤凰票据', 'MonteCarlo': 100.0397417322641, 'PDE': 100.0740834978835}
    expected_greeks = {'price_pde': 100.0740834978835, 'delta': 0.30693610759495726, 'gamma': -0.04191889361477763,
                       'vega': -0.6873103876611708, 'theta': 0.03410578625100413, 'rho': -0.3006235127738677}
    res_df, greeks = phoenix_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_standard_snowball_demo():
    expected_data = {'结构': '经典雪球(平敲雪球)', 'MonteCarlo': 99.99237077293571, 'PDE': 99.97242502284875,
                     'Quadrature': 100.01021095357333, 'pde_pct': '-0.020%', 'quad_pct': '0.018%'}
    expected_greeks = {'price_pde': 99.97242502284875, 'delta': 0.4336714279992435, 'gamma': -0.06160759915984215,
                       'vega': -0.7502448878742314, 'theta': 0.04425609512557571, 'rho': -0.19984183877359385}
    res_df, greeks = standard_snowball_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_earlyprofit_snowball_demo():
    expected_data = {'结构': '早利雪球', 'MonteCarlo': 99.97019888727107, 'PDE': 100.01918085274494,
                     'Quadrature': 100.04938425723334}
    expected_greeks = {'price_pde': 100.01918085274494, 'delta': 0.7485946436239885, 'gamma': -0.06779574769430496,
                       'vega': -0.4714328707591591, 'theta': 0.04273790484251094, 'rho': 0.03519575112444784}
    res_df, greeks = earlyprofit_snowball_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_butterfly_snowball_demo():
    expected_data = {'结构': '蝶式雪球', 'MonteCarlo': 100.00465809608768, 'PDE': 100.0364514590044,
                     'Quadrature': 100.0229173107232}
    expected_greeks = {'price_pde': 100.0364514590044, 'delta': 0.6003979584642778, 'gamma': -0.057819747897752904,
                       'vega': -0.6736305421984952, 'theta': 0.03749840960706763, 'rho': -0.08840143603079298}
    res_df, greeks = butterfly_snowball_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_otm_snowball_demo():
    expected_data = {'结构': 'OTM雪球', 'MonteCarlo': 100.00358496495079, 'PDE': 100.01671991322786,
                     'Quadrature': 100.0695778009394}
    expected_greeks = {'price_pde': 100.01671991322786, 'delta': 0.13769578938424587, 'gamma': -0.0216993090541564,
                       'vega': -0.38326463263842925, 'theta': 0.01884155036181312, 'rho': -0.4585777782421445}
    res_df, greeks = otm_snowball_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_floored_snowball_demo():
    expected_data = {'结构': '保底雪球(限损雪球，不追保雪球)', 'MonteCarlo': 100.04932829418274,
                     'PDE': 100.05319061780084, 'Quadrature': 100.08646854508989}
    expected_greeks = {'price_pde': 100.05319061780084, 'delta': 0.30476874786146624, 'gamma': -0.041764074317384825,
                       'vega': -0.49279151109935526, 'theta': 0.028830963019615297, 'rho': -0.2606915995998662}
    res_df, greeks = floored_snowball_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_stepdown_snowball_demo():
    expected_data = {'结构': '降敲雪球', 'MonteCarlo': 99.99000080300591, 'PDE': 99.99882667508119,
                     'Quadrature': 100.04927179943412}
    expected_greeks = {'price_pde': 99.99882667508119, 'delta': 0.42893445638513583, 'gamma': -0.05185686827974223,
                       'vega': -0.6690423157686496, 'theta': 0.03908593488756651, 'rho': -0.07847959340392663}
    res_df, greeks = stepdown_snowball_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_bothdown_snowball_demo():
    expected_data = {'结构': '双降雪球', 'MonteCarlo': 99.99734493746297, 'PDE': 100.03946242145815,
                     'Quadrature': 100.02280615164219}
    expected_greeks = {'price_pde': 100.03946242145815, 'delta': 0.6031430326171687, 'gamma': -0.05735408333077885,
                       'vega': -0.3797294860883085, 'theta': 0.04343410636937506, 'rho': 0.041062551228748134}
    res_df, greeks = bothdown_snowball_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_parachute_snowball_demo():
    expected_data = {'结构': '降落伞雪球', 'MonteCarlo': 100.00140677547931, 'PDE': 99.98244385775044,
                     'Quadrature': 100.0196889879925}
    expected_greeks = {'price_pde': 99.98244385775044, 'delta': 0.40278176156137846, 'gamma': -0.05003004622639651,
                       'vega': -0.6793836270392006, 'theta': 0.03305856903973847, 'rho': -0.1495028896361532}

    res_df, greeks = parachute_snowball_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_snowball_plus_demo():
    expected_data = {'结构': '看涨雪球(雪球增强，雪球plus)', 'MonteCarlo': 99.98093281705195, 'PDE': 99.99901087665904,
                     'Quadrature': 100.05768904941405}
    expected_greeks = {'price_pde': 99.99901087665904, 'delta': 0.654490097031001, 'gamma': -0.04116324325255505,
                       'vega': -0.605280941062901, 'theta': 0.029991508381016274, 'rho': -0.01271800889799124}

    res_df, greeks = snowball_plus_demo.run()
    assert expected_data == pytest.approx(res_df, rel=1e-10)
    assert expected_greeks == pytest.approx(greeks, rel=1e-10)


def test_paris_snowball_demo():
    expected_data = {'pv': 100.01742416459761, 'delta': 0.43278297470384075, 'gamma': -0.03718829098156107,
                     'vega': -0.8608643540902108, 'theta': 0.039641960435830015, 'rho': -0.08765889543886374,
                     '结构': '巴黎雪球'}
    res = paris_snowball_demo.run()
    assert expected_data == pytest.approx(res, rel=1e-10)
